<?php

// Uncomment for CodeIgniter
//if ( ! defined('BASEPATH')) exit('No direct script access allowed');

/**
 * PassHash
 * ========
 * 
 * This class hashes passwords before they are stored in a database.
 * Every password for account creation and login goes through here
 * to ensure that the hash is consistant. 
 * 
 * The password is not only salted, but also 'peppered' and a nonce is added
 * together with some md5 and sha1 hashing.
 * 
 * This mixture is then hash_hmac with sha512 and a site key is used. 
 * 
 * Complete usage example in the end of the file.
 *
 * The user of this class is encourage to change values for pepper, site key and hash template 
 * to ensure unique hashing.
 * 
 * --- Cracking ---
 * This hash can be cracked (or a rainbow table created) if the attacker
 *  - has access to the database record holding the password hash
 *  - has access to the database record holding the salt and the nonce uniquely assigned to the user. (Makes the use of rainbow tables completely impractical)
 *  - has access to the pepper and site key located on the file server (this file)
 *  - knows the correct algorithm/composition of password, salt, pepper and nonce 
 *
 * @author Peter Rasmussen (aka. Zadco) - php.peterrasmussen@spamgourmet.com
 * @link https://github.com/Zadco/PassHash
 */
class PassHash
{

    public function __construct($params = NULL)
    {
        
    }

    /**
     * Creates and returns an instance of PassHash_Instance class
     * That instance does all the work
     * The reason for this architecture is the way CodeIgniter manages libraries
     * 
     * @return PassHash_Instance
     */
    public function create()
    {
        return new PassHash_Instance();
    }

    /**
     * Destroys the instance of the PassHash_Instance class
     * It is first set to NULL then unset
     * Note: Unset does not actually unset the instance 
     * only unlinks the reference, but it does not hurt.
     * 
     * @param class $instance
     * @return boolean 
     */
    public function destroy(&$instance)
    {
        if ($instance instanceof PassHash_Instance)
        {
            $instance = NULL;
            unset($instance);
            return true;
        }
        else
            return false;
    }

}

/**
 * The instance made for every password hash
 * Create a new instance with
 * $instance = PassHash->create();
 * Destroy it again with 
 * PassHash->destroy(instance);
 */
class PassHash_Instance
{

    public function __construct($params = NULL)
    {
        
    }

    /**
     * Array containing password, salt and nonce
     * This array can be accessed with the designated methods ($this->password() returns $this->details['password'] etc.)
     * 
     * @var array 
     */
    private $details = array(
        "user_id" => NULL,
        "password" => NULL,
        "salt" => NULL,
        "nonce" => NULL
    );

    /**
     * Determine whether the current password is hashed or not
     * 
     * @var boolean 
     */
    private $password_hashed = false;

    /**
     * The template for hashing algorithm
     * 
     * @var string
     */
    private $hash_template = "[S] [P] sha1([PW] [N]) [PW] md5([PW] [S]) [N]";

    /**
     * Set the password
     * Note: The password can not be acquired,
     * only reset until it is hashed
     * 
     * @param string $password
     * @return boolean or string 
     */
    public function password($password = NULL)
    {
        if ($password == NULL && $this->password_hashed)
        {
            return $this->details['password'];
        }
        else if ($password !== NULL)
        {
            $this->password_hashed = false;
            return $this->details['password'] = $password;
        }
        else
        {
            return false;
        }
    }

    /**
     * Set or get user_id
     * User_id ensures unique hashing 
     * so a rainbow table only can be used for a specific user
     * - But salt and nonce takes care of that
     * 
     * @param string $user_id
     * @return boolean or string 
     */
    public function user_id($user_id = NULL)
    {
        if ($user_id == NULL)
        {
            return $this->details['user_id'];
        }
        else if ($user_id !== NULL)
        {
            return $this->details['user_id'] = $user_id;
        }
        else
        {
            return false;
        }
    }

    /**
     * Set or get the salt
     * If a string is passed the salt is set
     * else the current salt is returned 
     * 
     * @param string $salt
     * @return string or boolean
     */
    public function salt($salt = NULL)
    {
        if ($salt == NULL)
            return $this->details['salt'];
        else if ($salt !== NULL)
            return $this->details['salt'] = $salt;
        else
            return false;
    }

    /**
     * Set or get the nonce
     * If a string is passed the nonce is set
     * else the current nonce is returned
     * 
     * @param string $nonce
     * @return string or boolean
     */
    public function nonce($nonce = NULL)
    {
        if ($nonce == NULL)
            return $this->details['nonce'];
        else if ($nonce !== NULL)
            return $this->details['nonce'] = $nonce;
        else
            return false;
    }
    
    /**
     * Set or get the hash template
     * If a string is passed the nonce is set
     * else the current nonce is returned
     * 
     * @param string $template
     * @return string or boolean
     */
    public function template($template = NULL)
    {
        if ($template == NULL)
            return $this->hash_template;
        else if ($template !== NULL)
            return $this->hash_template = $template;
        else
            return false;
    }

    /**
     * Executes the password hashing
     * It does not return the hashed password,
     * it must be obtained by $this->get() 
     * 
     * @return boolean
     */
    public function hash()
    {
        if ($this->details['password'] == NULL)
            return false;

        //This hash is fairly strong without the user_id so it can be omitted
        if ($this->details['user_id'] == NULL)
            $this->details['user_id'] = "";

        if ($this->details['salt'] == NULL)
            $this->details['salt'] = $this->__salt();

        if ($this->details['nonce'] == NULL)
            $this->details['nonce'] = $this->__nonce();

        $pepper = $this->__pepper;
        $site_key = $this->__site_key;

        // Call to private hash function
        $password = $this->__hash(
                $this->details['password'], $this->details['salt'], $pepper, $this->details['nonce'], $site_key
        );

        $this->details['password'] = $password; // Override plaintext password with hashed password. 
        $this->password_hashed = true; // This allows you to access the newly hashed password

        return true;
    }

    /**
     * Returns the hashed password. 
     * This can only be done when the password is already hashed.
     * @return string or false
     */
    public function get()
    {
        if ($this->password_hashed)
        {
            return $this->details['password'];
        }
        else
        {
            return false;
        }
    }

    /* Private methods */

    /**
     * Hashes the plaintext password with this super secret algorithm
     * Feel free to play around with the line between 'sha512' and $site_key to get a unique algorithm
     * 
     * @param string $password
     * @param string $salt
     * @param string $pepper
     * @param string $nonce
     * @param string $site_key
     * @return string 
     */
    private function __hash($password, $salt, $pepper, $nonce, $site_key)
    {
        $data = array(
            "[PW]" => $password,
            "[S]" => $salt,
            "[P]" => $pepper,
            "[N]" => $nonce
        );

        $password_hash = hash_hmac(
                'sha512', $this->template_parser($data), $site_key
        );

        return $password_hash;
    }

    /**
     * Generates a 30 char long salt from a random number
     * 
     * @return string 
     */
    private function __salt()
    {
        $salt = "";

        $salt_length = 30;

        $salt = substr(md5(uniqid(rand(), true)), 0, $salt_length);

        return $salt;
    }

    /**
     * Unused method
     * Generates a string from array and string lookups from random numbers
     * 
     * @return string 
     */
    private function __token()
    {
        $token = "";
        $token_length = 32;
        $charset = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()";
        $words = array(
            "fluttershy", "crocodile", "gorm", "5k3Y", "Klondike", "USA", "qwerty",
            "bubber", "E=mc2", "SPARTA", "hitchhiker", "3.14159", "twilight sparkle", "iTunes",
            "å,.py", "æøå", "life", "minus", "waiting", "Verbatim", "notch",
            "rainbow dash", "minecraft", "42", "rarity", "occupy wallstreet", "applejack", "protect ip"
        );

        for ($i = 0; $i < $token_length; $i++)
        {
            if (rand(0, 10) > 1)
            {
                $randnum = rand(0, strlen($charset) - 1);
                $char = substr($charset, $randnum, 1);
            }
            else
            {
                $randnum = rand(0, count($words) - 1);
                $md5_word = md5($words[$randnum]);
                $char = substr($md5_word, rand(0, strlen($md5_word)), 1);
            }
            $token .= $char;
        }

        return $token;
    }

    /**
     * Generates a nonce from a hash of makeRandomString
     * 
     * @return string 
     */
    private function __nonce()
    {
        $nonce = 0;

        $nonce = hash('sha512', $this->__makeRandomString());

        return $nonce;
    }

    /**
     * Generates a random string from a number of bits
     * @author "ircmaxell"
     * @link http://stackoverflow.com/questions/4145531/how-to-create-and-use-nonces
     * 
     * @param int $bits
     * @return string 
     */
    private function __makeRandomString($bits = 256)
    {
        $bytes = ceil($bits / 8);
        $return = '';
        for ($i = 0; $i < $bytes; $i++)
        {
            $return .= chr(mt_rand(0, 255));
        }
        return $return;
    }

    /**
     * Parses the hash template and inserts designated values for password etc.
     * 
     * /!\ Buggy and unstable /!\
     * Leave template as-is to avoid errors
     * 
     * @param $source string
     * @param $token_table array
     * @return string
     */
    public function template_parser($token_table, $find = NULL, $template = NULL)
    {
        if ($token_table == NULL)
            return false;

        if ($find == NULL)
            $find = array("[S]", "[P]", "[PW]", "[N]", "md5(", "sha1(", ")");

        if ($template == NULL)
            $template = $this->hash_template;

        $chains = array();

        foreach ($find as $value)
        {
            $chains[$value] = array();
            for ($i = 0; $i < strlen($value); $i++)
            {
                $chains[$value][$value{$i}] = 0;
            }
        }

        // Point-in-chain
        $pic = array();

        $tokens = array();

        for ($i = 0, $input = $template . " "; $i < strlen($input); $i++)
        {
            $char = $input{$i};

            foreach ($chains as $chain_key => $chain)
            {
                foreach ($chain as $link_key => $link)
                {
                    if ($link == false)
                    {
                        $pic[$chain_key] = $link_key;
                        break;
                    }
                }
            }

            foreach ($chains as $chain_key => $chain)
            {

                if ($pic[$chain_key] == $char)
                {
                    $chains[$chain_key][$pic[$chain_key]] = 1;
                }
                else
                {
                    $this->reset_chain($chains, $chain_key);
                }

                if (count(array_unique(array_flip($chain))) == 1 && current($chain) != 0)
                {
                    array_push($tokens, "$chain_key");
                    $this->reset_chain($chains, $chain_key);
                }
            }
        }


        $output = "";

        $md5_following = false;
        $md5_string = "";
        $sha1_following = false;
        $sha1_string = "";

        foreach ($tokens as $token)
        {
            if ($token == "md5(")
            {
                $md5_following = true;
                continue;
            }
            if ($token == "sha1(")
            {
                $sha1_following = true;
                continue;
            }

            if ($md5_following && $token != ")")
            {
                $md5_string .= $token_table[$token];
            }
            if ($sha1_following && $token != ")")
            {
                $sha1_string .= $token_table[$token];
            }

            if ($token == ")" && $md5_following)
            {
                $md5_following = false;
                $output .= md5($md5_string);
                $md5_string = "";
                continue;
            }
            if ($token == ")" && $sha1_following)
            {
                $sha1_following = false;
                $output .= sha1($sha1_string);
                $sha1_string = "";
                continue;
            }

            if (!$md5_following && !$sha1_following)
                $output .= $token_table[$token];
        }

        return $output;
    }

    private function reset_chain(&$chains, $chain_key)
    {
        foreach ($chains[$chain_key] as $link_key => $link)
        {
            $chains[$chain_key][$link_key] = 0;
        }
    }

    /* Shhh! You haven't seen anything */

    private $__pepper = '$2a$07$rainbowdash4thAI/uJz89hj';
    private $__site_key = '}LB;i7&U^@27q%w+#]Z,%+44>,:wtg';

}

/**
 * Usage
 * =====
 * 
 * Initializing
 * ------------
 * 
 * // CodeIgniter
 * $this->load->library("PassHash");
 * $passhash = $this->passhash->create();
 * 
 * // Regular PHP
 * require_once("passhash.php");
 * $passhash = new PassHash()->create();
 * 
 * Account Creation
 * ----------------
 * 
 * // Hashing 
 * $passhash->password($input_data['password']);
 * $passhash->hash();
 * 
 * // Get hashed password, nonce & salt
 * $passhashed = $passhash->get(); OR $passhashed = $passhash->password();
 * $salt = $passhash->salt();
 * $nonce = $passhash->nonce();
 * 
 * // Store the 3 items above in a database
 * 
 * Login
 * -----
 * 
 * $passhash->salt($database_data->salt);
 * $passhash->nonce($database_data->nonce); 
 * $passhash->password($input_data['password']);
 * 
 * $passhash->hash();
 * 
 * if ($passhash->get() == $database_data->password)
 * {
 *     //Success
 * }
 */
?>
